using System;
using System.Runtime.Serialization;
using MT.MathBase;

namespace MT.DataStructures
{
	public enum eFPType
	{
		measured,
		calculated
	}

	public enum eFPLinkType
	{
		Zero			= 0,
		Backward		= 1,
		Forward			= 2,
		DoubleLinked	= 3
	}

	/// <summary>
	/// the FeaturePoint Class, stores a 2-Tupel for the X and Y axis of a Picture
	/// and the color of the Featurepoint.
	/// Additionally it knows to which Trajectory it belongs (tid)
	/// and which FeaturePoint of a frame it is (fid).
	/// 
	/// !!! Please note, that the fid of the FeaturePoint is !!!
	/// !!! _NOT_ the FrameID as used in other classes!      !!!
	/// 
	/// </summary>
	[Serializable]
	public class FeaturePoint: vector2f, IComparable //, ISerializable
	{

		//Connections to FP of other Frames
		// TODO: not sure if this is necessary
		private FeaturePoint _next = null;
		private	FeaturePoint _previous = null;

		//TrajectoryID of Featurepoint - untracked: id = -1, tracked: id > 1
		private int			_tid = -1;

		//Frame SubID of Featurepoint
		private int			_fid = -1;

		//Type of Featurepoint, has it ben measured or calculated
		private eFPType		_type = eFPType.measured;

		private eColorCode	_C = eColorCode.none;

		public	FeaturePoint() {}
		public	FeaturePoint(float X, float Y) : base(X,Y)
		{
			_C = eColorCode.none;
	
		}
		public  FeaturePoint(vector2f Coords, eColorCode Color): base(Coords)
		{
			_C = Color;
		}
		public 	FeaturePoint(float X, float Y, eColorCode Color) : base(X,Y)
		{
			_C = Color;
		}
		public 	FeaturePoint(float X, float Y, eColorCode Color, eFPType Type) : base(X,Y)
		{
			_C = Color;
			_type = Type;
		}
		public  FeaturePoint(vector2f Coords, eColorCode Color, int fid): base(Coords)
		{
			_C = Color;
			_fid =fid;
		}
		public 	FeaturePoint(float X, float Y, eColorCode Color, int fid) : base(X,Y)
		{
			_C = Color;
			_fid =fid;
		}
		/// <summary>
		/// Deserialization Constructor which recreates a featurepoint from
		/// a given Input String
		/// </summary>
		/// <param name="Input">String of a FeaturePoint which Serialised 
		/// itself by using its ToString() Method</param>
		public FeaturePoint(string Input)
		{
			string[] Data = new string[6];
			Data = Input.Split(":".ToCharArray(),6);

			_X = float.Parse(Data[0]);
			_Y = float.Parse(Data[1]);
			_tid = int.Parse(Data[2]);
			_fid = int.Parse(Data[3]);
			_type = (eFPType)int.Parse(Data[4]);
			_C = (eColorCode)int.Parse(Data[5]);

		}


		public	eColorCode	Color
		{
			get { return _C;  }
			set { _C = value; }
		}

		public	eFPLinkType  LinkType
		{
			get 
			{
					 if (_previous == null && _next == null) return eFPLinkType.Zero;
				else if (_previous == null && _next != null) return eFPLinkType.Forward;
				else if (_previous != null && _next == null) return eFPLinkType.Backward;
				else return eFPLinkType.DoubleLinked; 
			}
		}

		public	eFPType		Type
		{
			get
			{	return _type; }

		}

		/// <summary>
		/// Returns a link to the next FeaturePoint in this Trajectory - returns null if not linked yet.
		/// </summary>
		public	FeaturePoint next
		{
			get { return _next;  }
			set { _next = value; }
		}

		/// <summary>
		/// Returns a link to the previous FeaturePoint in this Trajectory - returns null if not linked yet.
		/// </summary>
		public	FeaturePoint previous
		{
			get { return _previous;  }
			set { _previous = value; }
		}

		/// <summary>
		/// Returns Trajectory ID of this Featurepoint - Value is -1 when untracked
		/// </summary>
		public	int			tid
		{
			get { return _tid;  }
			set { _tid = value; }
		}

		/// <summary>
		/// Returns Frame SubID of this Featurepoint
		/// </summary>
		public	int			fid
		{
			get { return _fid;  }
			set { _fid = value; }
		}

		public override string ToString()
		{
			string splitchar = ":";
			return string.Concat(new string[] {_X.ToString() ,splitchar, 
											   _Y.ToString(), splitchar, 
										       _tid.ToString(), splitchar, 
											   _fid.ToString(), splitchar, 
											   ((int)_type).ToString(), splitchar,
											   ((int)_C).ToString()}
								);
		}



		#region IComparable Members

		public int CompareTo(object obj)
		{

			// TODO:  Add FeaturePointComparer.Compare implementation
			FeaturePoint P1 = this;
			FeaturePoint P2 = (FeaturePoint)obj;

			if (P1.X == P2.X && P1.Y == P2.Y)
				return 0;
			else
			{
				if(P1.X*100000 + P1.Y > P2.X*100000 + P2.Y)
					return 1;
				else
					return -1;
			}
		
		}

		#endregion


/// Serialisation on FeaturePoint was tested and proved to be too slow for our requirements
/// therefor we created a Serialization Method by overriding the ToString() Method, which is
/// much faster than Serialization. 
/// For Deserialization look out for the Constructor that takes a string as argument.
///   
//		#region ISerializable Members
//		public void GetObjectData(SerializationInfo info, StreamingContext context)
//		{	//base.GetObjectData(info,context);
////			private FeaturePoint _next = null;
////			private	FeaturePoint _previous = null;
//
//			info.AddValue("_X", _X);
//			info.AddValue("_Y", _Y);
//
//			//TrajectoryID of Featurepoint - untracked: id = -1, tracked: id > 1
//			info.AddValue("_tid",_tid);
//
//			//Frame SubID of Featurepoint
//			info.AddValue("_fid",_tid);
//
//			//Type of Featurepoint, has it ben measured or calculated
//			//private eFPType		_type = eFPType.measured;
//			info.AddValue("_type",_type,_type.GetType());
//
//			//Color of Featurepoint
//			info.AddValue("_C",_C,_C.GetType());
//
//		}
//
//		private FeaturePoint(SerializationInfo info, StreamingContext context)
//		{	
//			_X = info.GetSingle("_X");
//			_Y = info.GetSingle("_Y");
//
//			_tid = info.GetInt32("_tid");
//			_fid = info.GetInt32("_fid");
//			_type = (eFPType)info.GetValue("_type",_type.GetType());
//			_C = (eColorCode)info.GetValue("_C",_C.GetType());
//		}
//
//	#endregion

	}

}
