using System;
using System.Collections;

namespace MT.DataStructures
{
	/// <summary>
	/// Stores Featurepoints and offers methods to access the points
	/// with names that are consistent to the IPAN Tracker nomenclature
	/// </summary>
	public class FeaturePointData:   IFeaturePointData, ICloneable
	{
		private	SortedList	mFrames = new SortedList();
		private string		_cid;

		public string Cid
		{
			get { return _cid; }
			set { _cid = value; }
		}

		public FeaturePointFrame LastFrame
		{
			get {	if (mFrames.Count>0)
						return (FeaturePointFrame)mFrames.GetByIndex(mFrames.Count-1); 
					else
						return null;}
		}
		public int LastFrameID
		{
			get 
			{
				if (mFrames.Count>0)
					return (int)mFrames.GetKey(mFrames.Count-1); 
				else
					return -1;
			}
		}

		public	FeaturePointData		(string cid)
		{	_cid = cid;
		}

		public	FeaturePointData		(FeaturePointFrame _Old, FeaturePointFrame _Cur, FeaturePointFrame _New)
		{
			AddFrame(1,_Old);
			AddFrame(2,_Cur);
			AddFrame(3,_New);
		}


		public	SortedList			FeaturePointFrames
		{
			get {return (SortedList)mFrames.SyncRoot;}
		}

		/// <summary>
		/// Used to store the loaded Featurepointsets into the FeaturePointData set.
		/// </summary>
		/// <param name="Number">Number of the frame</param>
		/// <param name="Points">Set of Points that belong to the frame</param>
		/// <returns></returns>
		public	bool				AddFrame	(int fnr, FeaturePointFrame Frame)
		{
			try		{ 
				Frame.cid = _cid;			//Set CameraID on each Frame
				mFrames.Add(fnr,Frame); }
			catch	{ return false; }

			return true;
		}


		public	FeaturePointFrame	GetFrame	(int fnr)
		{
			if (this.existsFrame(fnr))
			{	return ((FeaturePointFrame)mFrames[fnr]); }
			else
			{	return null;	}

		}

		public	void				RemoveFrame	(int fnr)
		{
			if (this.existsFrame(fnr))
			{	mFrames.Remove(fnr);
			}
		}

		public void					RemoveFirstFrame()
		{
			mFrames.RemoveAt(0);
		}


		public	FeaturePoint		P			(int Frame, int Index)
		{	
			if (this.existsFrame(Frame))
			{	return (FeaturePoint)((FeaturePointFrame)mFrames[Frame]).GetByIndex(Index); }
			else
			{	return null;	}
		}
		

		/// <summary>
		/// Checks if a certain frame has already been added to the Dataset
		/// </summary>
		/// <param name="Frame">Number of the Frame</param>
		/// <returns></returns>
		public	bool				existsFrame	(int Frame)
		{
			return mFrames.Contains(Frame);
		}


		/// <summary>
		/// Returns the number of Featurepoints contained in a given Frame
		/// </summary>
		/// <param name="Frame">Number of Frame</param>
		/// <returns></returns>
		public	int			Count		(int Frame)
		{
			if (mFrames.Contains(Frame))
			{
				return ((FeaturePointFrame)mFrames[Frame]).Count; 
			}
			else
			{
				return -1;	
			}

		}


		/// <summary>
		/// Returns the number of Frames contained in the Data Collection
		/// </summary>
		/// <returns></returns>
		public	int			CountFrames	()
		{
				return mFrames.Count;
		}
		#region ICloneable Member

		/// <summary>
		/// Creates a shallow Copy of FeaturePointData
		/// </summary>
		/// <returns></returns>
		public object Clone()
		{
			FeaturePointData Copy = new FeaturePointData(this._cid);
			Copy.mFrames = (SortedList)this.mFrames.Clone();

			return (object)Copy;
		}

		#endregion
	}

}
