using System;
using MT.DataStructures;
using MT.MathBase;

namespace MT.Tracker
{
	/// <summary>
	/// A Hypothesis object contains a Triplet of Featurepoints from 
	/// 3 consecutive Frames which fullfilled all constraints to be
	/// considered in further tracking analysis.
	/// </summary>
	public class Hypothesis
	{
		/** <summary>FeaturePoint of Frame 1</summary> */ 
		FeaturePoint _P1;	

		/** <summary>FeaturePoint of Frame 2</summary> */ 
		FeaturePoint _P2; 

		/** <summary>FeaturePoint of Frame 3</summary> */ 
		FeaturePoint _P3;	

		/** <summary>Stores the calculated Costs for this Hypothesis</summary> */ 
		float _Cost;

		/** <summary>Stores the Index of this Hypothesis</summary> */ 
		int	_Index;

		/** <summary>Stores the State of the Hypothesis, not used at the moment</summary> */ 
		eState _state;

		/** <summary>Unused, possible States of a Hypothesis: unverified, rejected, tested</summary> */ 
		public enum eState
		{	unverified = -1,
			rejected = 0,
			tested = 1
		}

		/// <summary>
		/// Hypothesis Constructor, takes 1 Featurepoint from 3 consecutive Frames,
		/// combined with the cost value which has been calculated by the tracker.
		/// </summary>
		/// <param name="P1">FeaturePoint of Frame 1</param>
		/// <param name="P2">FeaturePoint of Frame 2</param>
		/// <param name="P3">FeaturePoint of Frame 3</param>
		/// <param name="Cost">Cost of this Hypothesis</param>
		public Hypothesis(FeaturePoint P1, FeaturePoint P2, FeaturePoint P3,float Cost)
		{		
			_P1=P1;
			_P2=P2;
			_P3=P3;

			//Debug Code
			if (_P1.next !=null && _P2 != _P1.next)
			{
				Console.WriteLine("Invalid Hypothesis");
			}
			if (_P2.previous !=null && _P1 != _P2.previous)
			{
				Console.WriteLine("Invalid Hypothesis");
			}
			if (_P2.next !=null && _P3 != _P2.next)
			{
				Console.WriteLine("Invalid Hypothesis");
			}
			if (_P3.previous !=null && _P2 != _P3.previous)
			{
				Console.WriteLine("Invalid Hypothesis");
			}
			//End Debug Code
			
			_Index = Index;
			_Cost = Cost;

		}

		/// <summary>
		/// Tries to connect all Featurepoints of the Hypothesis and therefor
		/// creates a Trajectory out of them.
		/// Before connecting the FeaturePoints will be tested if they
		/// are already in other trajectories.
		/// </summary>
		/// <returns>True if connect() was succesful.</returns>
		public	bool	connect() //int TrajectorID
		{
			// x o o
			if (_P1.next == null && _P2.previous == null &&
				_P2.next == null && _P3.previous == null)
			{
				_P1.next = _P2;
				_P2.previous = _P1;
				_P2.next = _P3;
				_P3.previous = P2;

				return true;
			}
			// x-x o
			else if (_P1.next ==_P2 && _P2.previous == _P1 && 
				_P2.next == null && P3.previous == null)
			{
				_P2.next = _P3;
				_P3.previous = P2;
				
				return true;
			}
			// x y o <- must not happen
			else 
			{
				Console.WriteLine("Invalid Connect Attempt");
			}

			return false;

		}

		/// <summary>
		/// Returns the Angle between the Vectors which translate
		/// P1 -> P2 and P2 -> P3. This Value is used for the calculation
		/// of the Hypothesis Costs
		/// </summary>
		public	double		 Angle
		{
			get 
			{
				vector2f a = new vector2f(_P2.X - _P1.X, 
					_P2.Y - _P1.Y);

				vector2f b = new vector2f(_P3.X - _P2.X, 
					_P3.Y - _P2.Y);

				return Math.Acos(a*b/(a.length*b.length))/2/Math.PI*360;
			}
		}

		/// <summary>
		/// Used for Debug Output
		/// </summary>
		/// <returns></returns>
		public override string ToString()
		{
			return string.Format("H({0}): P({1,2})-P({2,2})-P({3,2}): {4}",_Index,P1.fid,P2.fid,P3.fid,_Cost.ToString("n:0.000000"));
		}


		#region "Properties"

		/** <summary>FeaturePoint of Frame 1</summary> */ 
		public	FeaturePoint P1
		{
			get { return _P1; }
		}

		/** <summary>FeaturePoint of Frame 2</summary> */ 
		public	FeaturePoint P2
		{
			get { return _P2; }
		}

		/** <summary>FeaturePoint of Frame 3</summary> */ 
		public	FeaturePoint P3
		{
			get { return _P3; }
		}

		/** <summary>Costs for this Hypothesis</summary> */ 
		public	float	Cost
		{	get	{ return _Cost; } 
		}

		/** <summary>Index of this Hypothesis</summary> */ 
		public  int		Index
		{
			get { return _Index;  }
			set { _Index = value; }
		}

		/** <summary>State of this Hypothesis</summary> */ 
		public	eState	State
		{
			get	{ return _state;  }
			set { _state = value; }
		}

		#endregion

	}
}




/// Legacy Code
/// contains some Deltafunctions for tracker Testing
/// 
//			vector2f a = new vector2f(_P2.X - _P1.X, 
//									  _P2.Y - _P1.Y);
//
//			vector2f b = new vector2f(_P3.X - _P2.X, 
//									  _P3.Y - _P2.Y);
//
//			w1 = (float)w1*a.length/Vmax;
//			w2 = 1- w1;
//
//			//Prevent Zero Values
//			double al = Math.Max(a.length,0.01);
//			double bl = Math.Max(b.length,0.01);
//
//			double alxbl = al * bl;
//
//			_Cost = (float)(w1*(1-Math.Max(a*b/alxbl,0)) + 
//				            w2*(1-2*Math.Sqrt(alxbl)/(al + bl)));

//			vector2f a = new vector2f(P2.X - P1.X, 
//									  P2.Y - P1.Y);
//
//			vector2f b = new vector2f(P3.X - P2.X,
//									  P3.Y - P2.Y);
//
//			w1 = 0.5f;
//			w2 = 0.5f;
//
//			double dx = Math.Abs(a.X - b.X);
//			double dy = Math.Abs(a.Y - b.Y);

//			_Cost =  (float)(w1*dx/Vmax + w2*dy/Vmax);

			//_Cost  = Gamma(_P1,_P2,_P3, Vmax);	


//		public static float Gamma(FeaturePoint P1, FeaturePoint P2, FeaturePoint P3, int Vmax)
//		{
//			float w1 = 0.9f;	//Angleweight
//			float w2 = 0.1f;	//Speedweight
//
//			vector2f a = new vector2f(_P2.X - _P1.X, 
//									  _P2.Y - _P1.Y);
//
//			vector2f b = new vector2f(_P3.X - _P2.X, 
//									  _P3.Y - _P2.Y);
//
//			w1 = (float)w1*a.length/Vmax;
//			w2 = 1- w1;
//
//			//Prevent Zero Values
//			double al = Math.Max(a.length,0.01);
//			double bl = Math.Max(b.length,0.01);
//
//			double alxbl = al * bl;
//
//			_Cost = (float)(w1*(1-Math.Max(a*b/alxbl,0)) + 
//				            w2*(1-2*Math.Sqrt(alxbl)/(al + bl)));
//				
//			vector2f a = new vector2f(P2.X - P1.X, 
//									  P2.Y - P1.Y);
//
//			vector2f b = new vector2f(P3.X - P2.X,
//									  P3.Y - P2.Y);
//
//			float w1 = 0.5f;
//			float w2 = 0.5f;
//
//			double dx = Math.Abs(a.X - b.X);
//			double dy = Math.Abs(a.Y - b.Y);
//
//			return (float)(w1*dx/_Vmax + w2*dy/_Vmax);

//		}

