using System;
using System.IO;
using MT.DataStructures;
using System.Collections;
using MT.MathBase;

namespace MT.Tracker
{
	/// <summary>
	/// Summary description for Tracker.
	/// </summary>
	public class Tracker: ITracker
	{

		
		//Events & Delegates
		public event NewHypothesisesDelegate NewHypothesises;

		private bool	_ConfigMode=false;	//Puts the tracker into a state where it raises events
									//whenever a HypothesisSet has been created

		//Tracker Settings
		private int		_Vmax = 40;			//Maximum allowed Speed in Pixels
		private int		_Vmin = 2;			//Minimal Speed so that AngleChanges are considered
		private int		_Vnrm = 10;			//Norm Speed from which on the normal weight in cost function is considered
		private float	_Gmax = 0.7f;		//Maximum allowed Costs

	    private int		_Vbase;				// = _Vnrm-_Vmin;

		//Gamma Settings
		private float _w1 = 0.9f;	//Angleweight
		private float _w2b = 0.1f;	//Speedweight
		private float _w2a = 0.0001f;



		//int		_Nver = 2;			//Verification Depth
		private string		_Color;

		private int			_TrajectorID = 0;

		SortedList _Trajectories=new SortedList();

		#region "Properties"

		public int Vmax
		{
			get {	return _Vmax;	}
			set {	_Vmax = value;	}
		}

		public int Vmin
		{
			get {	return _Vmin;	}
			set {	_Vmin = value; 
					calcVbase();
				}
		}

		public int Vnrm
		{
			get {	return _Vnrm;	}
			set {	_Vnrm = value; 
					calcVbase();
				}
		}

		public float GammaMax
		{
			get { return _Gmax;		}
			set { _Gmax = value;	}
		}

		public float AngleWeight
		{
			get { return _w1;		}
			set { _w1 = value;	}
		}

		public float SpeedWeight
		{
			get { return _w2b;		}
			set { _w2b = value;	}
		}

		public float StartSpeedWeight
		{
			get { return _w2a;		}
			set { _w2a = value;	}
		}

		public bool	 ConfigMode
		{
			get { return _ConfigMode; }
			set { _ConfigMode = value; }
		}

		#endregion 


		public Tracker(string color)
		{			
			calcVbase();
			_Color = color;
			//_Console = new StreamWriter("debug_" + color + "_"  + CameraID + ".txt");
		}
		public Tracker(bool ConfigMode, int Vmax, int Vmin, int Vnrm, float Gmax, float w1, float w2a, float w2b)
		{
			_ConfigMode = ConfigMode;
			_Vmax = Vmax;
			_Vmin = Vmin;
			_Vnrm = Vnrm;
			_Gmax = Gmax;
			_w1 = w1;
			_w2a = w2a;
			_w2b = w2b;

			calcVbase();

		}

		private void	calcVbase()
		{
				_Vbase = _Vnrm-_Vmin;
		}

		public bool initialize(FeaturePointFrameColor oldFrame, FeaturePointFrameColor curFrame, FeaturePointFrameColor newFrame)
		{
			float cost = 2; 

			//Check all 3 frames for existance - if one does not exist return false
			if (oldFrame==null || 
				curFrame==null || 
				newFrame==null)
			{	return false; }

			//_Console.WriteLine("****************** Initializing Track {0}/{1} *******************", newFrame.fid, curFrame.fid);

			HypothesisSuperSet HRank = new HypothesisSuperSet();

			

//			// Manually identify all points of the first frame
//			for(int j=0;j<oldFrame.Count;j++)
//			{
//				oldFrame.P(j).tid = j+1;
//			}
//
			//For each Point in Frame 2
			for(int i=0;i<curFrame.Count;i++)
			{
				for (int n=0;n<oldFrame.Count;n++)
				{
					//Check if P(k-1,n) is in Searchrange to P(k,i)
					if ((oldFrame.P(n)-curFrame.P(i)).length <= _Vmax) 
					{
						for (int m=0;m<newFrame.Count;m++)
						{
							//Check if P(k+1,m) is in Searchrange to P(k,i)
							if((newFrame.P(m)-curFrame.P(i)).length <= _Vmax)
							{
								cost = gamma(oldFrame.P(n),curFrame.P(i),newFrame.P(m));
								
								if (cost<= _Gmax)
								{
									//Create new Hypothesis
									HRank.Add(new Hypothesis(oldFrame.P(n),
															 curFrame.P(i),
															 newFrame.P(m),
											                 cost));

									//_Console.WriteLine("H({7}): P({0},{1}) - P({2},{3}) - P({4},{5}) : {6}",oldFrame.fid,n,curFrame.fid,i,newFrame.fid,m,cost, H.Index);
								}
							}			
						}
					}
				}
			}

			//Announce found Hypothesises if activated
			if (_ConfigMode)
				if (NewHypothesises!= null)
					NewHypothesises(HRank);

//			int oldTrajectorID = _TrajectorID;

			HypothesisSet HBest = HRank.getBestSet();

			foreach (Hypothesis H in HBest.Values)
			{
				//Tries to connect the Hypothesis, throw exception if not successful
				if (H.connect()) //++_TrajectorID
				{
					//Legacy Code - intented to store and debug trajectories
//					//Store Starting point of new Trajectory
//					if (_TrajectorID !=oldTrajectorID)
//					{
//						_Trajectories.Add(_TrajectorID,H.P1);
//						oldTrajectorID = _TrajectorID;
//					}
					//_Console.WriteLine("H({7}): {8:f} |-T-> {0}: P({1},{2}) - P({3},{4}) - P({5},{6})",H.P1.tid, H.P1.X, H.P1.Y,H.P2.X, H.P2.Y,H.P3.X, H.P3.Y,H.Index,Math.Round(H.Angle,2));
				}
				else 
					throw new Exception("Tracker: Could not connect Hypothesis");

			}

			return true;

		}
	
		public bool processDoubleFrame(FeaturePointFrameColor curFrame, FeaturePointFrameColor newFrame)
		{


			//Verify existence of required Frames
			if(curFrame == null ||
			   newFrame == null)
			{	return false;	}

			//_Console.WriteLine("****************** Processing Double Frame {0}/{1} *******************", newFrame.fid, curFrame.fid);

			int				i;
			int				j;
			float[,]		error = new float[curFrame.Count,newFrame.Count];
			FeaturePoint	P1;
			FeaturePoint	P2;

			//Double Processing requires at least 1 FP in each Frame
			//else there is nothing to do at all
			if (curFrame.Count > 0 && newFrame.Count > 0)
			{
				//Initialize errorMin Array
				for(i = 0;i< curFrame.Count;i++)
					for(j = 0;j< newFrame.Count;j++)
					{
						error[i,j] = 1000000;
					}


				//For each FP in k-1 find its closest Equivalent in the new frame identical FPs
				for(i = 0;i< curFrame.Count;i++)
				{
					P1 = curFrame.P(i);

					for(j = 0;j<newFrame.Count;j++)
					{
						P2 = newFrame.P(j);

						error[i,j] = (P1.X -P2.X) * (P1.X -P2.X) + (P1.Y -P2.Y) * (P1.Y -P2.Y);

					}

				}

				float minimum;
				int minI = -1;
				int minJ = -1;

				do
				{
					minimum = 1000000;

					for(i = 0;i< curFrame.Count;i++)
					{
						for(j = 0;j<newFrame.Count;j++)
						{
							if (error[i,j]<minimum)
							{
								minimum =error[i,j];
								minI = i;
								minJ = j;
							}
						}
					}

					//If there are still FPs to connect, do that
					if (minimum != 1000000)
					{
						P1 = curFrame.P(minI);
						P2 = newFrame.P(minJ);
						P1.next = P2;
						P2.previous = P1;
//						P2.tid = P1.tid; // Done by Postprocessing
						//_Console.WriteLine(" ->T-> {0}: P({1},{2}) - P({3},{4})",P1.tid, P1.X, P1.Y,P2.X, P2.Y);

					}

					//Delete used points from the error Matrix
					//by setting the errorvalues to 1000000
					for(i = 0;i< curFrame.Count;i++)
						error[i,minJ] = 1000000;

					for(j = 0;j<newFrame.Count;j++)
						error[minI,j] = 1000000;


				}while (minimum!= 1000000);
		
			}

			return true;
		}

		public bool processFrame(FeaturePointFrameColor oldFrame, FeaturePointFrameColor curFrame, FeaturePointFrameColor newFrame)
		{
			float cost = 2; 

			HypothesisSuperSet HRank = new HypothesisSuperSet();
			
			
			//_Console.WriteLine("********************** Processing Frames {0}-{1}-{2} **********************", oldFrame.fid, curFrame.fid,newFrame.fid);

			//Verify that k > 1 and necessary Frames are existing
			// ToDo: Deactivated Initializer Test because there is no way to confirm it
			//if (!(k > 1)) return false;
			if (oldFrame==null || 
				curFrame==null || 
				newFrame==null)
			{	return false; }

			//For each Point in curFrame
			for(int i=0;i<curFrame.Count;i++)
			{
				if (curFrame.P(i).LinkType == eFPLinkType.Backward )
				{	
					for (int m=0;m<newFrame.Count;m++)
					{
						//Check if P(k+1,m) is in Searchrange to P(k,i)
						if((newFrame.P(m)-curFrame.P(i)).length <= _Vmax)
						{							
							cost = gamma(curFrame.P(i).previous,curFrame.P(i),newFrame.P(m));
							if (cost<= _Gmax)
							{
								//Create new Hypothesis from a continued trajectory
								HRank.Add(new Hypothesis(curFrame.P(i).previous,
														 curFrame.P(i),
														 newFrame.P(m),
														 cost));

								//_Console.WriteLine("H({7}): P({0},{1}) - P({2},{3}) - P({4},{5}) : {6}",oldFrame.fid,curFrame.P(i).previous.fid,curFrame.fid,i,newFrame.fid,m,H.Cost, H.Index);
							}
						}			
					}				
				}
				//Action to take if P(k,i) is not backward linked
				else
				{
					for (int n=0;n<oldFrame.Count;n++)
					{
						//Check if P(k-1,n) is in Searchrange to P(k,i)
						//and that it is not yet forward linked to another P(k,j)
						if (((oldFrame.P(n)-curFrame.P(i)).length <= _Vmax) && (oldFrame.P(n).next == null)) 
						{
							for (int m=0;m<newFrame.Count;m++)
							{
								//Check if P(k+1,m) is in Searchrange to P(k,i)
								if((newFrame.P(m)-curFrame.P(i)).length <= _Vmax)
								{
									cost = gamma(oldFrame.P(n),curFrame.P(i),newFrame.P(m));
									if (cost<= _Gmax)
									{
										//Create new Hypothesis
										HRank.Add(new Hypothesis(oldFrame.P(n),
																 curFrame.P(i),
																 newFrame.P(m),
																 cost));
									
										//_Console.WriteLine("H({7}): P({0},{1}) - P({2},{3}) - P({4},{5}) : {6}",oldFrame.fid,n,curFrame.fid,i,newFrame.fid,m,cost, H.Index);

									}
								}			
							}
						}
					}
				}
			}

			//Announce found Hypothesises if activated
			if (_ConfigMode)
				if (NewHypothesises!= null)
					NewHypothesises(HRank);

			HypothesisSet HBest = HRank.getBestSet();	
						
			foreach (Hypothesis H in HBest.Values)
			{
				//Tries to connect the Hypothesis
				if (H.connect()) //++_TrajectorID
				{
					//if (H.P1.LinkType == FeaturePoint.eFPLinkType.DoubleLinked)
						//_Console.WriteLine("H({7}): {8:f} ->T-> {0}: P({1},{2}) - P({3},{4}) - P({5},{6})",H.P1.tid, H.P1.X, H.P1.Y,H.P2.X, H.P2.Y,H.P3.X, H.P3.Y,H.Index,H.Angle);
					//else
						//_Console.WriteLine("H({7}): {8:f} |-T-> {0}: P({1},{2}) - P({3},{4}) - P({5},{6})",H.P1.tid, H.P1.X, H.P1.Y,H.P2.X, H.P2.Y,H.P3.X, H.P3.Y,H.Index,Math.Round(H.Angle,2));

				}
				else 
					throw new Exception("Tracker: Could not connect Hypothesis");
			}

			return true;
		}


		public bool postProcessFrame(FeaturePointFrameColor oldFrame)
		{
			if(oldFrame !=null)
			{
//				FeaturePoint P;

				foreach(FeaturePoint P in oldFrame.Values)
				{
					if (P.tid == -1 && P.previous == null)
						P.tid = ++_TrajectorID;
					else if(P.tid==-1 && P.previous != null)
						P.tid = P.previous.tid;
				}


//				for(int i = 0; i<oldFrame.Count;i++)
//				{
//					P = oldFrame.P(i);
//					if(P.tid ==-1 && P.previous == null)
//						P.tid = ++_TrajectorID;
//					else if(P.tid==-1 && P.previous != null)
//						P.tid = P.previous.tid;
//					
//				}

			}

			return true;

		}

		private float velocity(FeaturePoint P1, FeaturePoint P2)
		{ return (float)Math.Sqrt(Math.Pow((P2.X-P1.X),2)+ Math.Pow((P2.Y-P1.Y),2));
		}

		// TODO: The costfunction gamma is currently implemented twice, one time in the Tracker class and a second time in the Hypothesis class. One of them is going to be redundant later on.
		private float gamma(FeaturePoint P1, FeaturePoint P2, FeaturePoint P3)
		{	
			vector2f a = new vector2f(P2.X - P1.X, 
									  P2.Y - P1.Y);

			vector2f b = new vector2f(P3.X - P2.X,
			                          P3.Y - P2.Y);

			
			float v = Math.Max(a.length,b.length);
			
			double axb = a.length * b.length;	//abs(a)*abs(b)
			double aub = a.length + b.length;	//abs(a)+abs(b)

			float w1, w2;

			if (v <_Vmin)
			{
				w1 = 0;

				return (float)(_w2a*(1-2*Math.Sqrt(axb)/(Math.Max(aub,0.01))));
			}
			else if (v <_Vnrm)
			{   //Fit Weightfactors
				w1 = _w1*(v-_Vmin)/(_Vbase);
				w2 = (_w2b-_w2a)*(v-_Vmin)/(_Vbase);
//					Math.Min(wmax-_w1,_w2a);
			}
			else
			{
				w1 = _w1;
				w2 = _w2b;
			}

			return (float)(w1*(1-Math.Max(a*b/axb,0)) + w2*(1-2*Math.Sqrt(axb)/(aub)));
			//return (float)(w1*(0.5-0.5*Math.Max(a*b/axb,0)) + w2*(1-2*Math.Sqrt(axb)/(aub)));

		
		}

//		private float	gamma(FeaturePoint P1, FeaturePoint P2, FeaturePoint P3)
//		{
//			
//			vector2f a = new vector2f(P2.X - P1.X, 
//				P2.Y - P1.Y);
//
//			vector2f b = new vector2f(P3.X - P2.X,
//				P3.Y - P2.Y);
//
//			float w1 = 0.5f;
//			float w2 = 0.5f;
//
//			double dx = Math.Abs(a.X - b.X);
//			double dy = Math.Abs(a.Y - b.Y);
//
//			return (float)(w1*dx/_Vmax + w2*dy/_Vmax);
//
//		}
		
		
		public void storeTrajectories()
		{
			//Try to open File
			// TODO: Open File to Store Trajectories for Debug Reasons
			//try { myreader = new StreamReader("trajectories.csv"); }
			//catch { _Console.WriteLine("ERROR: Could not open File \"" + mstrPath + mstrName + fnr.ToString(strMaxDigits) + "." + mstrEnding + "\""); }		

		}

		public void Finish()
		{
			//_Console.Flush();
			//_Console.Close();
			
		}
		}
}
