using System;
using tsai;
using MT.DataStructures;
using System.Collections;

namespace MTServer
{



	/// <summary>
	/// The TUKController is the main tracking Class in the TUK System.
	/// It only operates on Data of one color Type and can therefor only use
	/// Trajectory and gemoetric calculated Data independend of colors to 
	/// do a successful tracking.
	/// It contains TUKElements which each correspond to one single realWorld object.
	/// -> one table tennis ball in our version.
	/// </summary>
	public class TUKController
	{

		public event TrackerInformationDelegate TrackerInformation;

		/// <summary>
		/// Information string to send DebugInformation from TUKElements to 
		/// parental Objects by usung the TrackerInformation event.
		/// </summary>
		string				_strInfo="";

		/// <summary>
		/// Contains a List of all Cameras that are currently connected to the server.
		/// </summary>
		clsCameras			_Cameras;

		/// <summary>
		/// Contains a list of the Bodypoint Data of one color. This Datastructure
		/// contains all the 3D results which will be calculated during the tracking
		/// process
		/// </summary>
		clsBPDataColor		_BPDataColor;

		/// <summary>
		/// Contains List of all TUKElemnts of one color - a TUKElement corresponds to a single tracked Element
		/// in the Real World - which is a lighted Ball in our case.
		/// </summary>
		SortedList			_TUKElements = new SortedList();

		/// <summary>
		/// Stores all Trajectory,TUKElement Relations per camera.
		/// 
		/// This Array stores all Trajectory IDs from each Camera combined with their association 
		/// to their TUKElement. The Tracking process relies on this trajectory information
		/// to associated Featurepoints which belong to a known Trajectory very fast with the correct
		/// TUKElement
		/// </summary>
		Hashtable[]			_trajectories;					


		/** <summary>Contains the Modelsize</summary> */
		float		_size				= 165;

		/** <summary>Contains the Color of this Controller</summary> */
		private eColorCode _color;


		#region "AssociationProperties"
		public float Size
		{
			get {return _size; }
			set {_size = value;}
		}

		public clsCameras ClientCollection
		{
			get {return _Cameras; }
		}
		#endregion


	/// <summary>
	/// Creates a TUKController Object which controlls the tracking process inside the TUKElements.
	/// Since we use colors to improve the tracking, a TUKElement only works on one single Colortype.
	/// </summary>
	/// <param name="cameras">Contains all connected Cameras</param>
	/// <param name="BPDataColor">Contains the Datastructure in which the tracking output is stored</param>
		public TUKController(clsCameras cameras, clsBPDataColor BPDataColor)
		{
			_BPDataColor = BPDataColor;
			_color = _BPDataColor.Color;
			_Cameras = cameras;
		
			ResetElements();
		}

		/// <summary>
		/// Creates a TUK Element Structure to allow Tracking of the currently loaded BPScheme
		/// -> a BPScheme contains a predefined structure of the expected tracking model -
		/// for example a human body with 13 Bodypoints or a bar with only 2 Bodypoints.
		/// This Function has to be called each time when a new BPScheme has been loaded. 
		/// </summary>
		public void ResetElements()
		{
			_TUKElements.Clear();

			TUKElement TE;

			foreach (clsBPStream BPStream in _BPDataColor.Values)
			{
				//Create new Element and add it to local hashtable
				TE = new TUKElement(BPStream,_Cameras);
				_TUKElements.Add(TE.Name, TE);

				//Connect Debugout Events
				TE.TrackerInformation +=new TrackerInformationDelegate(TE_TrackerInformation);

			}
		}

		/// <summary>
		/// This is the Main Tracking Procedure which will execute all tracking
		/// related actions to tracke one single Frame.
		/// </summary>
		/// <param name="Frames">A Collection of synchronised Frames from each Camera</param>
		public void TrackFrame(clsFPFrameArrayColor Frames)
		{
			_strInfo = "";

			//Foodanalogy
			//-----------
			//This procedure is shown as an analogy of eaters = TUKElements
			//which are hungry and want to eat food = Featurepoints
			//Each TUKElement wants to eat one Featurepoint from each Camera
			//as long as it didn't get fed with a Featurepoint from a certain 
			//camera it will be hungry.

			//For our Datamanagement, we create a Hungry-Hashtable for each Camera
			//in which we initally store all TUKElement.
			//During the Algorithm we will start to remove the TE whenever they are
			//fed
			FeaturePoint P;
			TUKElement TE;
			SortedList[] HungryTEs = new SortedList[_Cameras.Count];
			SortedList[] Food = new SortedList[_Cameras.Count];


			int curFrame=Frames.fid;
	
			//Prepare list of Food and table of Hungry Bodypoints
			for (int c=0;c<_Cameras.Count;c++)
			{

				//Assertion(curFrame must NEVER be equal or lower than 0)
				if (curFrame<=0){throw new Exception("Kat.1 AAAARGH");}

				HungryTEs[c] = new SortedList(_TUKElements);

				Food[c] = new SortedList((SortedList)Frames.GetByIndex(c));
				
			}

			//Assertion(curFrame must NEVER be equal or lower than 0)
			if (curFrame<=0){throw new Exception("Kat.3 AAAARGH");}

			//for (int c=0;c<_Cameras.Count;c++)
			foreach(FeaturePointFrameColor Frame in Frames.Values)
			{
				int c = Frames.IndexOfValue(Frame);

				//First Feed the FPs which belong to 
				//known Trajectories
				for(int i = 0;i<Frame.Count;i++)
				{	
					P = Frame.P(i);
					
					if (_trajectories[c].Contains(P.tid))
					{
						TE = (TUKElement)_trajectories[c][P.tid];
						TE.Feed(curFrame,c,P);
						HungryTEs[c].Remove(TE.Name);
						Food[c].Remove(P.fid);
					}

				}

				//If only one hungry TE and one FP is left, feed it to that TE
				//This action might have to be checked for consistency
				if(HungryTEs[c].Count==1 && Food[c].Count ==1)
				{
					//Get last TUKlement
					IEnumerator lastTE = HungryTEs[c].Values.GetEnumerator();
					lastTE.MoveNext();
					TE = (TUKElement)lastTE.Current;

					//Get last Featurepoint
					IEnumerator lastFP = Food[c].Values.GetEnumerator();
					lastFP.MoveNext();
					P = (FeaturePoint)lastFP.Current;

					TE.Feed(curFrame,c,P);
					HungryTEs[c].Remove(TE.Name);
					Food[c].Remove(P.fid);

					//Store new Trajectory in Trajectories list
					_trajectories[c].Add(TE.LastKnownPosition(c).tid,TE);

				}
			}

			//After all trajectory related FPs have been fed - calculate first temporary position
			//to allow assoziation of newly started trajectories and detection of errors
			foreach(DictionaryEntry DE in _TUKElements)
			{
				TE = (TUKElement)DE.Value;
				TE.calc(curFrame);

				int cid = TE.getWorstCamera();
				if(cid!=-1)
				{
					//Remove Featurepoint of a camera which gave us bad intersections -> it is most propably wrong
					P = TE.Unfeed(curFrame,cid);
					_trajectories[cid].Remove(P.tid);
					Food[cid].Add(P.fid,P);
					HungryTEs[cid].Add(TE.Name,TE);

//					Console.WriteLine("Removed bad FP with Trajectory {0} from Camera {1}.",P.tid, _Cameras.GetByIndex(cid).Name);
				}

			}

			double v1 = 0;				
			double v2 = 0;

			for (int c=0;c<_Cameras.Count;c++)
			{
				if (HungryTEs[c].Count > 0 && Food[c].Count>0)
				{

					//Create Distance Array
					float[,] Distance = new float[HungryTEs[c].Count,Food[c].Count];
					int i;
					int j;
					FeaturePoint TE_FP;

					//Search the best Food for the hungry Bodypoints
					//which only require one more food
					for(i=0; i<HungryTEs[c].Count;i++)
					{
						TE = (TUKElement)HungryTEs[c].GetByIndex(i);
						TE_FP = TE.get2DCoord(curFrame,c);

						//If the TE has been detected by at least 2 cameras, try to associate it with 
						//the unassociated FPs of the non-tracking camera
						if (TE_FP !=null)
						{	
							for(j=0;j<Food[c].Count;j++)
							{
								P = (FeaturePoint)Food[c].GetByIndex(j);
						
								//Calculate Squared Distance in Pixels
								v1 = (TE_FP.X-P.X);				
								v2 = (TE_FP.Y-P.Y);
								Distance[i,j] = (float)(v1*v1+v2*v2);
							}
						}
						else 
						{
							for (j = 0;j<Food[c].Count;j++)
							{
								Distance[i,j]=1000000;
							}
						}


					}

					float minimum;
					int minI = -1;
					int minJ = -1;

					do
					{
						minimum = 100000000;

						for(i = 0;i<HungryTEs[c].Count;i++)
						{
							for(j = 0;j<Food[c].Count;j++)
							{
								if (Distance[i,j]<=minimum)
								{
									minimum =Distance[i,j];
									minI = i;
									minJ = j;
								}
							}
						}

						//If a valid minimum has been found,
						//check if the distance is lower than the required minimal distance
						//for continuing trajectories.
						//If this is the case, feed the FP to the TE
						if (minimum != 100000000)
						if (minimum < 40)
						{
							TE = (TUKElement)HungryTEs[c].GetByIndex(minI);
							P = (FeaturePoint)Food[c].GetByIndex(minJ);
							
							//Deactivated, network code doesnt restore connection between featurepoints
							//Only assoziate brand new trajectories since we don't want crossing trajectories to connect
//							if(P.LinkType == eFPLinkType.Forward)
//							{
								TE.Feed(curFrame,c,P);
								HungryTEs[c].Remove(TE.Name);
								Food[c].Remove(P.fid);

								//Store new Trajectory in Trajectories list
								_trajectories[c].Add(TE.LastKnownPosition(c).tid,TE);

//								Console.WriteLine("Frame {3}: connected new Trajectory {2} from Camera {0} to Bodypoint {1}",c+1,TE.name,P.tid,curFrame);
//							}
						}

						//Delete used points from the error Matrix
						//by setting the errorvalues to 1000000
						for(i = 0;i< HungryTEs[c].Count;i++)
							Distance[i,minJ] = 100000000;

						for(j = 0;j<Food[c].Count;j++)
							Distance[minI,j] = 100000000;


					}while (minimum!= 100000000);


				}
			}


			for (int c=0;c<_Cameras.Count;c++)
			{
			}

			foreach(DictionaryEntry DE in _TUKElements)
			{
				TE = (TUKElement)DE.Value;
//				if(TE.name == "Fuss_Rechts")
//				{
//					Console.WriteLine("iQuality von {0} betraegt {1}", TE.name,TE.intersectionQuality(curFrame));
//				}
			}



			//After all FPs have been fed - commit definite positions
			foreach(DictionaryEntry DE in _TUKElements)
			{
				TE = (TUKElement)DE.Value;
				TE.commit(curFrame);
				
//				if(TE.name == "Fuss_Rechts")
//					Console.WriteLine("mQuality: {0}", TE.moveQuality(curFrame));

			}

			OnTrackerInformation(_strInfo);

		}

/*
 						//If a valid BP Position could not be precalced (the BP has only been seen
						//by 1 camera for some time) use the last known position to determine 
						//its distance
						if (BP_FP== null) { BP_FP = (BP.LastKnownPosition(c)); }
*/


		/// <summary>
		/// Eventhandler to collect all Debuginformation which is send from the 
		/// TUKElements
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void TE_TrackerInformation(object sender, TrackerEventArgs e)
		{
			_strInfo += e.Info;
		}

		/// <summary>
		/// EventRaiser Function to inform upper Objects about DebugInformation from all TUKELements
		/// and this TUKController
		/// </summary>
		/// <param name="strInfo"></param>
		public void OnTrackerInformation(string strInfo)
		{
			if (TrackerInformation!=null){TrackerInformation(this, new TrackerEventArgs(_strInfo));}
		}

		/// <summary>
		/// Assicociate tries to connect each FP of a Camera with the predefined Model
		/// which is stored in Frame 0 of the _BPData Structure. This Function is essential
		/// to begin with the tracking process and has to be called to initialise the tracking.
		/// 
		/// ToDo: There exists the Idea to use the association inbetween the tracking
		/// process to recollect lost Bodypoints. This could improve the tracking resultws
		/// by a great deal.
		/// </summary>
		/// <param name="Frames">Contains a FeaturePoint Frame from each connected Camera, which will be 
		/// assiciated with the Standard Model stored in Frame 0 of the _BPDataColor Structure</param>
		public void Associate(clsFPFrameArrayColor Frames)
		{
			int i;												//Counter Variable

			clsBPFrame BPFrameModel = _BPDataColor.getFrame(0);	//Stores 3-Dimensional FrameModel
			FeaturePointFrameColor FrameModel;					//Stores projected FrameModel for a Camera

			FeaturePointFrameColor Frame;						//Stores Frame of one Camera

			//Initialise _trajectories Hashtable Array
			_trajectories = new Hashtable[_Cameras.Count];

			for (i = 0; i< _Cameras.Count;i++)
			{
				_trajectories[i] = new Hashtable();
			}

			//Associate for each Camera
			for (int c=0;c<_Cameras.Count;c++)
			{
				clsCamera Camera = _Cameras.GetByIndex(c);

				// Verify existance of a Frame and that the camera is calibrated
				if (Frames[Camera.Name].Count!=0 && Camera.Tsai!=null)
				{
					//Calculate the Modelprojection for this Camera
					FrameModel = Resize3Dmodel(BPFrameModel,Camera);

					//Get the Frame of this Camera
					Frame = Frames[Camera.Name];
		
					//Associate the Frame of this Camera with its ModelFrame
					AssociateCamera(FrameModel, Frame, c);					
				}
			}
		}

		/// <summary>
		/// Associates a FrameModel with a given Frame by using a Least Squared Distance Algorithm
		/// -> a Distance Array between all FrameModel and Frame points is created and the ones with 
		/// the smalles distance are associated.
		/// The association happens in 2 Ways: 
		/// 1. The Model Featurepoints are directly connected with their associated Framepoints with
		///    the .next Property.
		/// 2. The Trajectory ID of the Frame FeaturePoint is stored in _trajectories together with 
		///    its associated TUKElement.
		/// </summary>
		/// <param name="FrameModel">Contains a predefined Model of the expected Structure</param>
		/// <param name="Frame">Contains the Featurepoints for the Frame that will be associated</param>
		/// <param name="cid">Contains the CameraID </param>
		/// <returns>Returns true if successfully esecuted</returns>
		public bool AssociateCamera(FeaturePointFrameColor FrameModel, FeaturePointFrameColor Frame, int cid)
		{
			const long distMax = 9999999999;

			//Verify existence of required Frames
			if(FrameModel == null ||
				Frame == null)
			{	return false;	}


			int				i;
			int				j;
			float[,]		dist = new float[FrameModel.Count,Frame.Count];
			FeaturePoint	P1;
			FeaturePoint	P2;

			//Double Processing requires at least 1 FP in each Frame
			//else there is nothing to do at all
			if (FrameModel.Count > 0 && Frame.Count > 0)
			{
				//Initialize distMin Array
				for(i = 0;i< FrameModel.Count;i++)
					for(j = 0;j< Frame.Count;j++)
					{
						dist[i,j] = distMax;
					}


				//For each FP in k-1 find its closest Equivalent in the new frame identical FPs
				for(i = 0;i< FrameModel.Count;i++)
				{
					P1 = FrameModel.P(i);

					for(j = 0;j<Frame.Count;j++)
					{
						P2 = Frame.P(j);

						dist[i,j] = (P1.X -P2.X) * (P1.X -P2.X) + (P1.Y -P2.Y) * (P1.Y -P2.Y);

					}

				}

				float minimum;
				int minI = -1;
				int minJ = -1;

				do
				{
					minimum = distMax;

					for(i = 0;i< FrameModel.Count;i++)
					{
						for(j = 0;j<Frame.Count;j++)
						{
							if (dist[i,j]<minimum)
							{
								minimum =dist[i,j];
								minI = i;
								minJ = j;
							}
						}
					}

					//If there are still FPs to associate, do that
					if (minimum != distMax)
					{
						//WP: we know that P2 is associated to P1
						P1 = FrameModel.P(minI);
						P2 = Frame.P(minJ);
						P1.next = P2;
						P2.previous = P1;
						P1.tid = P2.tid;

						TUKElement TE = (TUKElement)_TUKElements.GetByIndex(minI);

						if (P2.tid!=-1)
							_trajectories[cid].Add(P2.tid,TE);
			
					}

					//Remove associated points from the distance Matrix
					//by setting the distvalues to distMax
					for(i = 0;i< FrameModel.Count;i++)
						dist[i,minJ] = distMax;

					for(j = 0;j<Frame.Count;j++)
						dist[minI,j] = distMax;


				}while (minimum!= distMax);
		
			}

			return true;
		}

		/// <summary>
		/// Calculates the 3DModel of a predefinde BPScheme - since we use a human Model which
		/// depends on the size of the tracked human, the Model is calculated by using a given 
		/// size of the Human to track. The standard Scheme is defined for a human with height
		/// of 1.80m and therefor we scale according to this value.
		/// </summary>
		/// <param name="BPFrame"></param>
		/// <param name="Camera"></param>
		/// <returns></returns>
		public FeaturePointFrame Resize3Dmodel(clsBPFrame BPFrame, clsCamera Camera)
		{
			float factor = _size/180.0f;
			
			FeaturePointFrame FrameModel = new FeaturePointFrame(BPFrame.FrameID, Camera.Name);

			if (Camera.Tsai!=null)
			{
				int i=0;		//Index garantiert korrekte Reihenfolge innerhalb der SortedList FPFrameColor

				//This is a little Hack to enable us to use a FeaturePointFrame as return value.
				//Originally the FeaturePointFrame structure was not created for this kind of data,
				//and expectes different FeaturePoint SubframeIDs to work.
				//To prevent collisions with Data from othercolored Controllers,
				//we distribute the possible IDs so far apart that no collisions should happen
				//this hack prevents us from having more than 1000 Featurepoints, but I dont think
				//That will happen so soon :)
				// ToDo: find a better Method to solve this problem!
				switch(_color)
				{
					case eColorCode.R:
						i = 0;
						break;
					case eColorCode.G:
						i = 1000;
						break;
					case eColorCode.B:
						i = 2000;
						break;
				}

				foreach(BodyPoint BP in BPFrame.Values)
				{
					FrameModel.Add(new FeaturePoint(Camera.world_coord_to_image_coord(BP*factor),BP.Color, i++));
				}
			}

			return FrameModel;

		}

		/// <summary>
		/// Calculated 3DModels for all Cameras 
		/// </summary>
		/// <returns></returns>
		public clsFPFrameArray Resize3Dmodel()
		{
			//The Model is always stored in Frame 0 -> therefor framenr = 0
			clsBPFrame BPFrame = _BPDataColor.getFrame(0);
			clsFPFrameArray result = new clsFPFrameArray(0);

			foreach (clsCamera camera in _Cameras.Values)
			{
				result.Add(camera.Name, Resize3Dmodel(BPFrame, camera));
			}
			return result;
		}

	}

	
}
