using System;
using tsai;
using System.Collections;
using OpenGL;
using MT.MathBase;
using MT.DataStructures;

namespace MTServer
{
	/// <summary>
	/// The TUKMain Class is the 3D Tracking Controller, which operates on the 3 different basic
	/// Colors that we use in the Tracking System. It instantates the 3 TUK Controllers which each
	/// work on only one Color. It also seperates all incoming Data by this 3 Colors and works as 
	/// abstraction Layer for this colors. 
	/// Basically TUKMain does everything that a TUKController does, just combined for all RGB Colors
	/// instead of only one Color.
	/// </summary>
	public class TUKMain
	{

		public event TrackerInformationDelegate TrackerInformation;

		private TUKController		_TUK_R;
		private TUKController		_TUK_G;
		private TUKController		_TUK_B;


		private Hashtable			_bodypoints = new Hashtable();
		private Hashtable			_bodyconnectors = new Hashtable();


		private clsCameras			_Cameras;
		private clsBPData			_BPData;

		private string			_strInfo="";

		private bool				_associated = false;		//True after calling associate


		public delegate void NewTrackDelegate(int FrameID);
		public event NewTrackDelegate NewTrack;
		public event EventHandler Associated;
		public event FrameDelegate NewFrameModel; 


		//AssociationRGB Related Variables
		float		_size				= 165;

		int _selectedClient = 0;

		#region "AssociationRGB Properties"
		public float Size
		{
			get {return _size; }
			set 
			{
				_size = value;
				_TUK_R.Size= value;
				_TUK_G.Size= value;
				_TUK_B.Size= value;

				Resize3Dmodel();

			}
		}


		public FeaturePointFrame getModel2D(int frameID)
		{
			ICamera Camera = _Cameras.GetByIndex(_selectedClient);
			FeaturePointFrame frame = new FeaturePointFrame(-1,Camera.Name);
			int Count=0;
			float factor = _size/180.0f;

			clsBPFrame BPFrame = _BPData.getFrame(frameID);

			foreach(BodyPoint BP in BPFrame.Values)
			{
				vector2f vec = Camera.world_coord_to_image_coord(BP*factor);
				FeaturePoint FP = new FeaturePoint(vec,BP.Color, Count++);
				frame.Add(FP);
				
			}
			return frame; 

		}

		public clsCameras ClientCollection
		{
			get {return _Cameras; }
		}
		#endregion

		public TUKMain(clsCameras cameras, clsBPData BPData)
		{
			_Cameras = cameras;
			_BPData = BPData;

			_BPData.Loaded += new EventHandler(_BPData_Loaded);

			//Initalize 3 Trackers each for one color
			_TUK_R = new TUKController(_Cameras, _BPData.DataRed);
			_TUK_R.TrackerInformation+=new TrackerInformationDelegate(TUK_TrackerInformation);

			_TUK_G = new TUKController(_Cameras, _BPData.DataGreen);
			_TUK_G.TrackerInformation+=new TrackerInformationDelegate(TUK_TrackerInformation);

			_TUK_B = new TUKController(_Cameras, _BPData.DataBlue);
			_TUK_B.TrackerInformation+=new TrackerInformationDelegate(TUK_TrackerInformation);


		}

		// ToDo: Implement Loading of Bodypoint Schemes
		public bool LoadScheme()
		{
			return false;
		}

		public void Reset()
		{
			_TUK_R = new TUKController(_Cameras, _BPData.DataRed);
			_TUK_G = new TUKController(_Cameras, _BPData.DataGreen);
			_TUK_B = new TUKController(_Cameras, _BPData.DataBlue);

			Resize3Dmodel();
		}


//		public void CreateBodyPoint(string name, eColorCode color, FeaturePoint[] FPs)
//		{
//			TUKElement BP;
//
//
//			switch (color)
//			{
//				case eColorCode.R:
//					BP = new TUKElement(name, color, _Cameras);
//					_TUK_R.AddBodyPoint(name,BP,FPs);
//					_bodypoints.Add(name,BP);
//					break;
//				case eColorCode.G:
//					BP = new TUKElement(name, color, _Cameras);
//					_TUK_G.AddBodyPoint(name,BP,FPs);
//					_bodypoints.Add(name,BP);
//					break;
//				case eColorCode.B:
//					BP = new TUKElement(name, color, _Cameras);
//					_TUK_B.AddBodyPoint(name,BP,FPs);
//					_bodypoints.Add(name,BP);
//					break;
//			}
//		}

		/// <summary>
		/// Creates a Body Connector Object that connects 2 existing Bodypoints when they are painted.
		/// </summary>
		/// <param name="name">Name of the BodyConnector (must be unique to all BodyConnectors and Bodypoints or an exception is thrown</param>
		/// <param name="LeftConnectedBP">Name of the left connected Bodypoint</param>
		/// <param name="RightconnectedBP">Name of the right connected BOdypoint</param>
		public void CreateBodyConnector(string name, string LeftConnectedBP, string RightConnectedBP, float elasticity)
		{	
			_bodyconnectors.Add(name, new BodyConnector(name,(TUKElement)_bodypoints[LeftConnectedBP],(TUKElement)_bodypoints[RightConnectedBP],elasticity));
		}
		public void CreateBodyConnector(string name, string LeftConnectedBP, string FirstRightConnectedBP, string SecondRightConnectedBP, float elasticity)
		{	
			_bodyconnectors.Add(name, new BodyConnector(name,(TUKElement)_bodypoints[LeftConnectedBP],(TUKElement)_bodypoints[FirstRightConnectedBP],(TUKElement)_bodypoints[SecondRightConnectedBP],elasticity));
		}
		public void CreateBodyConnector(string name, string FirstLeftConnectedBP, string SecondLeftConnectedBP, string FirstRightConnectedBP, string SecondRightConnectedBP,float elasticity)
		{	
			_bodyconnectors.Add(name, new BodyConnector(name,(TUKElement)_bodypoints[FirstLeftConnectedBP],(TUKElement)_bodypoints[SecondLeftConnectedBP],(TUKElement)_bodypoints[FirstRightConnectedBP],(TUKElement)_bodypoints[SecondRightConnectedBP],elasticity));
		}
		/// <summary>
		/// Called by the parent, each time when a new Frame is available.
		/// </summary>
		public void TrackFrame(clsFPFrameArray Frame)
		{
			if(!_associated)
				Associate(Frame);

			_strInfo = "";

			_TUK_R.TrackFrame(Frame.F_Red);
			_TUK_G.TrackFrame(Frame.F_Green);
			_TUK_B.TrackFrame(Frame.F_Blue);


//			TUKElement BP;
//			int curFrame = _A[0].CountFrames()-3;
//			foreach(DictionaryEntry DE in _bodypoints)
//			{
//				BP = (TUKElement)DE.Value;
//
//				if(BP.name == "Huefte_Rechts")
//					BP.quality(curFrame);
////					Console.WriteLine("Die Connector Qualitaet von {0} betraegt {1}", BP.name,BP.quality(curFrame));
//			}

			if (NewTrack!=null)
				NewTrack(Frame.fid);

			OnTrackerInformation(_strInfo);


		}

		public void OnTrackerInformation(string strInfo)
		{
			if (TrackerInformation!=null){TrackerInformation(this, new TrackerEventArgs(_strInfo));}
		}


		public void PaintFrame(int FrameID, bool Leinwand)
		{
			// Weltkoordinaten Offset
			float dx = -2.4f;
			float dy = -1.2f;
			float dz =  0.0f;
			
			//Draw Floor
			GL.glColor3f(0.5f,0.5f,0.5f);
			for (int x=0; x<7; x++)
			{
				for (int y=0; y<4; y++)
				{
					WGL.drawRectangle(	new vector3f( (float)(x*1.2+1.18)+dx, (float)(y*1.2+1.18)+dy, 0.0f+dz),
										new vector3f( (float)(x*1.2+1.18)+dx, (float)(y*1.2+0.02)+dy, 0.0f+dz),
										new vector3f( (float)(x*1.2+0.02)+dx, (float)(y*1.2+1.18)+dy, 0.0f+dz),
										new vector3f( (float)(x*1.2+0.02)+dx, (float)(y*1.2+0.02)+dy, 0.0f+dz));
				}
			}
			WGL.drawRectangle(	new vector3f(10.2f+dx, -0.02f+dy, 0.0f+dz),
								new vector3f(10.2f+dx, -2.0f+dy,  0.0f+dz),
								new vector3f(-2.0f+dx, -0.02f+dy, 0.0f+dz),
								new vector3f(-2.0f+dx, -2.0f+dy,  0.0f+dz));

			WGL.drawRectangle(	new vector3f(10.2f+dx, 5.6f+dy,  0.0f+dz),
								new vector3f(10.2f+dx, 3.62f+dy, 0.0f+dz),
								new vector3f(-2.0f+dx, 5.6f+dy,  0.0f+dz),
								new vector3f(-2.0f+dx, 3.62f+dy, 0.0f+dz));

			WGL.drawRectangle(	new vector3f(10.2f+dx, 3.62f+dy,  0.0f+dz),
								new vector3f(10.2f+dx, -0.02f+dy, 0.0f+dz),
								new vector3f(7.22f+dx, 3.62f+dy,  0.0f+dz),
								new vector3f(7.22f+dx, -0.02f+dy, 0.0f+dz));

			WGL.drawRectangle(	new vector3f(-0.02f+dx, 3.62f+dy,  0.0f+dz),
								new vector3f(-0.02f+dx, -0.02f+dy, 0.0f+dz),
								new vector3f(-2.0f+dx,  3.62f+dy,  0.0f+dz),
								new vector3f(-2.0f+dx,  -0.02f+dy, 0.0f+dz));


			if (Leinwand)
			{

				foreach (clsCamera camera in _Cameras.Values)
				{
					camera.Paint();
				}
				// 2D-Bilder der einzelnen Kameras (Featurepoints)
//				vector3f center = new vector3f(3.6f,  1.8f,  1.0f);
//				for (int i=0; i<_Cameras.Length; i++)
//				{

					

					// Sichtpyramide
//					double xw = new double();
//					double yw = new double();
//					double zw = new double();
//
//					_Cameras[i].get_camera_position(ref xw, ref yw, ref zw);
//					vector3f campos = new vector3f((float)xw/1000, (float)yw/1000, (float)zw/1000);
//
//						
//					
//					GL.glColor3f(0,1,0);
//					_Cameras[i].image_coord_to_world_coord(0, 0, 0.01f, ref xw, ref yw);
//					vector3f dir1 = new vector3f((float)xw/1000, (float)yw/1000, 0.01f);
//					
//					_Cameras[i].image_coord_to_world_coord(639, 0, 0.01f, ref xw, ref yw);
//					vector3f dir2 = new vector3f((float)xw/1000, (float)yw/1000, 0.01f);
//
//					_Cameras[i].image_coord_to_world_coord(0, 479, 0.01f, ref xw, ref yw);
//					vector3f dir3 = new vector3f((float)xw/1000, (float)yw/1000, 0.01f);
//
//					_Cameras[i].image_coord_to_world_coord(639, 479, 0.01f, ref xw, ref yw);
//					vector3f dir4 = new vector3f((float)xw/1000, (float)yw/1000, 0.01f);
//		
//					
//					WGL.drawLine(campos, dir1);
//					WGL.drawLine(campos, dir2);
//					WGL.drawLine(campos, dir3);
//					WGL.drawLine(campos, dir4);
//
//					WGL.drawLine(dir1, dir2);
//					WGL.drawLine(dir2, dir4);
//					WGL.drawLine(dir4, dir3);
//					WGL.drawLine(dir3, dir1);
//
//
//					// Webcam
//					GL.glColor3f(0,0,0);
//
//					GL.glTranslatef(campos.X, campos.Y, campos.Z);
//					GL.glRotatef((float)(-_Cameras[i]._cc.Rx/Math.PI*180),1,0,0);
//					GL.glRotatef((float)(-_Cameras[i]._cc.Ry/Math.PI*180),0,1,0);
//					GL.glRotatef((float)(-_Cameras[i]._cc.Rz/Math.PI*180),0,0,1);
//					
//					GL.glRotatef(90,0,0,1);
//					WGL.drawSphere(0.1f, 16, 16, new vector3f(0.4f,0.4f,0.4f));
//					WGL.drawCylinder(0.03f, new vector3f(0,0,0), new vector3f(0.15f,0,0),36,new vector3f(0,0,0));
//					WGL.fillCircleYZ(new vector3f(0.15f,0,0),0.1f,36);
//					GL.glRotatef(-90,0,0,1);
//					GL.glRotatef(-90,0,1,0);
//					WGL.drawCylinder(0.03f, new vector3f(0,0,0), new vector3f(0.12f,0,0),36,new vector3f(0,0,0));
//					GL.glRotatef(90,0,1,0);
//
//					GL.glRotatef((float)(_Cameras[i]._cc.Rz/Math.PI*180),0,0,1);
//					GL.glRotatef((float)(_Cameras[i]._cc.Ry/Math.PI*180),0,1,0);
//					GL.glRotatef((float)(_Cameras[i]._cc.Rx/Math.PI*180),1,0,0);					
//					GL.glTranslatef(-campos.X, -campos.Y, -campos.Z);



					// Leinwand
//					GL.glTranslatef(_cameracoords[i].X/1000 + 2*(center.X-_cameracoords[i].X/1000),
//						_cameracoords[i].Y/1000 + 2*(center.Y-_cameracoords[i].Y/1000),
//						_cameracoords[i].Z/1000 + 2*(center.Z-_cameracoords[i].Z/1000));
//					GL.glRotatef(alpha,0,0,1);
//
//					//Rahmen der Leinwand
//					WGL.drawRectangle(	new vector3f(0,  0.02f, -0.48f),
//						new vector3f(0,  0.02f, -0.65f),
//						new vector3f(0, -0.02f, -0.48f),
//						new vector3f(0, -0.02f, -0.65f));
//
//					WGL.drawRectangle(	new vector3f(0,  0.02f, -0.65f),
//						new vector3f(0,  0.32f, -0.82f),
//						new vector3f(0, -0.02f, -0.65f),
//						new vector3f(0,  0.36f, -0.82f));
//
//					WGL.drawRectangle(	new vector3f(0,  0.02f, -0.65f),
//						new vector3f(0, -0.36f, -0.82f),
//						new vector3f(0, -0.02f, -0.65f),
//						new vector3f(0, -0.32f,	-0.82f));
//
//					WGL.drawRectangle(	new vector3f(0,  0.64f, -0.48f),
//						new vector3f(0, -0.64f, -0.48f),
//						new vector3f(0,  0.64f,  0.48f),
//						new vector3f(0, -0.64f,  0.48f));
//					// Leinwand
//					GL.glColor3f(1,1,1);
//					WGL.drawRectangle(	new vector3f(0.001f,  0.60f, -0.44f),
//						new vector3f(0.001f, -0.60f, -0.44f),
//						new vector3f(0.001f,  0.60f,  0.44f),
//						new vector3f(0.001f, -0.60f,  0.44f));
//
//				
////					for (int j=0; j<_A[i].Count(FrameID); j++ )
////					{
////						GL.glTranslatef(0.002f, (320-(640-_A[i].P(FrameID, j).X))/500 , -(_A[i].P(FrameID, j).Y/500 - 0.48f));
////						if(_A[i].P(FrameID, j).Color == eColorCode.R) GL.glColor3f(1,0,0);
////						if(_A[i].P(FrameID, j).Color == eColorCode.G) GL.glColor3f(0,1,0);
////						if(_A[i].P(FrameID, j).Color == eColorCode.B) GL.glColor3f(0,0,1);
////						WGL.fillCircleYZ(new vector3f(0,0,0), 0.03f, 12);
////						//WGL.drawSphere(0.03f, 12, 12);
////						GL.glTranslatef(-0.002f, -(320-(640-_A[i].P(FrameID, j).X))/500, (_A[i].P(FrameID, j).Y/500) - 0.48f);
////					}
//
//					GL.glRotatef(-alpha,0,0,1);
//					GL.glTranslatef(-(_cameracoords[i].X/1000 + 2*(center.X-_cameracoords[i].X/1000)),
//						-(_cameracoords[i].Y/1000 + 2*(center.Y-_cameracoords[i].Y/1000)),
//						-(_cameracoords[i].Z/1000 + 2*(center.Z-_cameracoords[i].Z/1000)));
				
//				}

				

//				foreach (clsClient Client in _ClientCollection.Values)
//				{
//					FeaturePointFrame FPframe = (FeaturePointFrame) Client.FPData.FeaturePointFrames.GetByIndex(FrameID);
//					foreach (FeaturePoint p in FPframe.Values)
//					{
//						GL.glTranslatef(0.002f, (320-(640-p.X))/500 , -(p.Y/500 - 0.48f));
//					
//						if(FPframe.Color == eColorCode.R) GL.glColor3f(1,0,0);
//						if(FPframe.Color == eColorCode.G) GL.glColor3f(0,1,0);
//						if(FPframe.Color == eColorCode.B) GL.glColor3f(0,0,1);
//						WGL.fillCircleYZ(new vector3f(0,0,0), 0.03f, 12);
//
//					
//						GL.glTranslatef(-0.002f, -(320-(640-p.X))/500, (p.Y/500) - 0.48f);
//					}
//				}
			}
//			GL.glColor3f(1,0,0);
//			_TUK_R.PaintFrame(FrameID);
//			GL.glColor3f(0,1,0);
//			_TUK_G.PaintFrame(FrameID);
//			GL.glColor3f(0,0,1);
//			_TUK_B.PaintFrame(FrameID);

//			TUKElement BP;
//			foreach(DictionaryEntry DE in _bodypoints)
//			{	
//				BP = (TUKElement)DE.Value;
//				BP.paint(FrameID);
//			}
			

			_BPData.Paint(FrameID);
//			BodyConnector BC;
//			foreach(DictionaryEntry DE in _bodyconnectors)
//			{	
//				BC = (BodyConnector)DE.Value;
//				BC.Paint(FrameID);
//			}
		}

		private void TUK_TrackerInformation(object sender, TrackerEventArgs e)
		{
			_strInfo += e.Info + "\n";
		}


		private void Associate(clsFPFrameArray Frame)
		{
			_TUK_R.Associate(Frame.F_Red);
			_TUK_G.Associate(Frame.F_Green);
			_TUK_B.Associate(Frame.F_Blue);

			OnAssociation();
		}

		public void OnAssociation()
		{
			_associated = true;
			if (Associated!=null){Associated(this, new EventArgs());}
		}
		

		public void Resize3Dmodel()
		{
			clsFPFrameArray frame = new clsFPFrameArray(0);

			frame.Add(_TUK_R.Resize3Dmodel());
			frame.Add(_TUK_G.Resize3Dmodel());
			frame.Add(_TUK_B.Resize3Dmodel());
			
			OnNewFrameModel(frame);

		}

		private void _BPData_Loaded(object sender, EventArgs e)
		{
			Reset();
		}

		public void OnNewFrameModel(clsFPFrameArray frame)
		{
			if (NewFrameModel != null)
			{
				NewFrameModel(this, new FrameEventArgs(frame));
			}
		}
	}
}
