using System;
using System.Net.Sockets;
using System.Net;
using System.Text;
using System.IO;
using System.Collections;
using System.Threading;
using System.Runtime.Serialization.Formatters.Binary;
using MT.Synchronization;

namespace MT.Network
{
/// <summary>
/// Server
/// This class provides a simple Serversystem. Setting a port in the constructor
/// and calling StartServer() will start a server. When a new connection is created, the server
/// will raise the NewConnection Event, announcing the new connected Client.
/// From that point on, the Server will also announce all events, that a Client sends.
/// Use Send() to send either String Messages or clsNet Objects to all clients. 
/// 
/// clsNet Objects are predefined
/// Object Container Classes which are serialised and sent to the clients
/// The connection can be closed
/// </summary>
	public class clsNetServer: SyncInvoker, IDisposable
	{
		//declare Event Delegates
//		public delegate void NewConnectionDelegate(clsNetClient NewClient);
//		public delegate void ClosedConnectionDelegate(string conID);
//		public delegate void NewMessageDelegate(string conID, string message);
//		public delegate void NewMessageObjectDelegate(string conID, clsNet message);
//		public delegate void NewErrorMessageDelegate(string conID, string errormessage);

		//declare Events
		public event NetClientHandler NewConnection ;
		public event NetEventHandler ClosedConnection;
		public event NetMessageHandler NewMessage;
		public event NetObjectHandler NewMessageObject;
		public event NetMessageHandler NewErrorMessage;

		protected SortedList	_Clients = new SortedList();
		protected TcpListener	_Listener;
		protected Thread		_ListenerThread;

		//Parameter for ReceiveLooper - must be reset to "" after being used!
		protected string	_conID="";
		protected int		_port;

		public clsNetServer(int Port)
		{
			_port = Port;

		}

		/// <summary>
		/// Starts the Listening Process of the Server. Calling this Function
		/// will enable Clients to connect to the server.
		/// </summary>
		public virtual void StartServer()
		{
			//TcpListener is listening on the given port...                                                 
			_Listener = new TcpListener(System.Net.IPAddress.Any, _port);
			_Listener.Start();    

			
			_ListenerThread = new Thread(new ThreadStart(Listener));
			_ListenerThread.IsBackground = true;

			_ListenerThread.Name = "Listener";
			_ListenerThread.Start();
			Thread.Sleep(0);
		}

		//Contains ListenerLoop which checks for new connection requests.
		//Every time a new connection request arrives it checks its conID
		//and creates and starts a Thread for the ReceiverLoop.
		private void Listener()
		{

			TcpClient		incomingCon;
			NetworkStream	netStream;
			clsNetNewCon	NewCon;
			clsNetClient		newClient;

			//used to deserialize incoming Objects
			BinaryFormatter bf = new BinaryFormatter();


			//Enter Message Loop
			while (true)
			{
				//AcceptTcpClient is a blocking Function which only continues,
				//when a new Connection Request arrives.
				incomingCon = _Listener.AcceptTcpClient();
				netStream = incomingCon.GetStream();

				clsNet msg = bf.Deserialize(netStream) as clsNet;
				//The first NetworkMessage of a new connection must always contain a ConnectionRequest Object
				//otherwise the Connecton is denied.
				NewCon = msg as clsNetNewCon;

				if (NewCon == null)	//First Packet of new Connection was no Connection Request -> Deny Request
				{
					bf.Serialize(netStream,new clsNetError("","Request Denied!",0));
					incomingCon.Close();
				}



				//If ConID is already used close the new connection - else create new Receiver Loop
				if (_Clients.ContainsKey(NewCon.ConID))
				{
					//Add a Connection Number to the ConID and return it to the client
					NewCon.ConID = NewCon.ConID + "_" + _Clients.Count;

					//					bf.Serialize(netStream,new clsNetError("","Request Denied! ConnectionID was already taken.",1));				
					//					incomingCon.Close();}
				}
				
				bf.Serialize(netStream, NewCon);

				newClient = new clsNetClient(NewCon, incomingCon);

				//Connect all Client Events to the server Events
				//newClient.NewConnection += new Network.clsNetClient.NewConnectionDelegate(OnNewConnection);
				newClient.ClosedConnection +=new NetEventHandler(OnClosedConnection);
				newClient.NewMessage += new NetMessageHandler(OnNewMessage);
				newClient.NewErrorMessage +=new NetMessageHandler(OnNewErrorMessage);
				newClient.NewMessageObject+= new NetObjectHandler(OnNewMessageObject);

				//Add Client to our Client Pool
				_Clients.Add(NewCon.ConID, newClient);
				

				//Announce new connection by raising event
				OnNewConnection(newClient);	
			}
		}

		private void OnNewConnection(clsNetClient NewClient)
		{	base.SyncInvoke(NewConnection,this,new NetClientEventArgs(NewClient));
			//if (NewConnection!=null){NewConnection(this, new NetClientEventArgs(NewClient));}
		}

		private void OnClosedConnection(object sender, NetEventArgs e)
		{	
			_Clients.Remove(e.ConID);
			//if (ClosedConnection!=null){ClosedConnection(sender, e);} 
			base.SyncInvoke(ClosedConnection,sender,e);
		}

		private void OnNewMessage(object sender, NetMessageEventArgs e)
		{	base.SyncInvoke(NewMessage,sender,e);
			//if (NewMessage!=null){NewMessage(sender, e);}
		}

		private void OnNewMessageObject(object sender, NetObjectEventArgs e)
		{	base.SyncInvoke(NewMessageObject,sender, e);
			//if (NewMessageObject!=null){NewMessageObject(sender, e);}
		}

		private void OnNewErrorMessage(object sender, NetMessageEventArgs e)
		{	base.SyncInvoke(NewErrorMessage,sender,e);
				//if (NewErrorMessage!=null){NewErrorMessage(sender, e);}
		}



		/// <summary>
		/// Disconnects all connected Clients and additionaly shuts down listener
		/// </summary>
		public void CloseConnection()
		{
			foreach(clsNetClient Client in _Clients.Values)
			{
				Client.CloseConnection();
				if (ClosedConnection != null) {ClosedConnection(this, new NetEventArgs(Client.ConnectionID));}
			}

			_Clients.Clear();

		}
		
		/// <summary>
		/// Disconnects specified Client
		/// </summary>
		/// <param name="conID"></param>
		public void CloseConnection(string conID)
		{
			//Shutdown the given Client
			if (_Clients.Contains(conID))
			{
				clsNetClient Client = (clsNetClient)_Clients[conID];
				Client.CloseConnection();
				_Clients.Remove(conID);
			
				//if (ClosedConnection != null) {ClosedConnection(conID);}
			}
		}


		/// <summary>
		/// Sends the specified Messagestring to all connected Clients. 
		/// </summary>
		/// <param name="message">Message to be sent.</param>
		/// <returns>Returns 'true' if successful, else 'false'.</returns>
		public bool Send(string message)
		{
			bool r = true;
			BinaryFormatter bf = new BinaryFormatter();

			foreach(clsNetClient Client in _Clients.Values)
			{
					r = r && Client.Send(message);
			}

			return r;
		}
		/// <summary>
		/// Sends the specified clsNet derived Object to all connected Clients
		/// </summary>
		/// <param name="message"></param>
		/// <returns></returns>
		public bool Send(clsNet message)
		{
			bool r = true;
			BinaryFormatter bf = new BinaryFormatter();

			foreach(clsNetClient Client in _Clients.Values)
			{
				r = r && Client.Send(message);
			}

			return r;

		}

		#region IDisposable Members

		public virtual void Dispose()
		{
			clsNetClient Client;

			//Stop Listening
			_Listener.Stop();
			_Listener = null;

			//Abort ListenerThread Loop
			_ListenerThread.Abort();
			Console.WriteLine("Aborted Thread " + _ListenerThread.Name) ;      

	
			//Close all Connections
			while (_Clients.Count!=0)
			{	Client = (clsNetClient)_Clients.GetByIndex(0);
				Client.CloseConnection();
				Client.Dispose();
			}

		}

		#endregion
	}
}
