// SWFilter.h: interface for the CSWFilter class.
// 
//////////////////////////////////////////////////////////////////////
//  CSWFilter: (A DirectShow (DirectX 8.0) filter)
//
//  Purpose: This filter is a basic Transform Filter to process video
//        
//  Usage: 
//  1.  Register the filter. 
//      regsvr32 SWFilter.ax
//
//  2.  Insert the filter into the graph to process video data (live 
//      capture or video files)
//      a.  Use the GraphEdt.exe (in DirectX SDK) to build a graph 
//      b.  To use this filter in App, include iSWFilter.h for
//          definition of interfaces and build the graph in the App.
///////////////////////////////////////////////////////////////////////
#pragma once

//Network Includes
#include "net/SocketException.h"
#include "net/ClientSocket.h"
#include <iostream>
#include <string>

#include <fstream>
#include "iSWFilter.h"
//#include "midlswfilter.h"

using namespace std;

class CSWFilter
    : public CTransformFilter,      // Derive from Transform Filter
      public ISWFilter,				// Interface of CSWFilter
      public ISpecifyPropertyPages, // Having property page
      public CPersistStream         // Persist the object
{
public:

    DECLARE_IUNKNOWN;

// Constructor && Deconstructor
    static CUnknown *WINAPI CreateInstance(LPUNKNOWN punk, HRESULT *phr);
    CSWFilter(TCHAR *tszName, LPUNKNOWN punk, HRESULT *phr);
    ~CSWFilter();

// Overrriden from CTransformFilter base class
    // Check if the transform type can be accomplished
    HRESULT CheckInputType(const CMediaType* mtIn);
    HRESULT CheckTransform(const CMediaType *mtIn, const CMediaType *mtOut);
    HRESULT DecideBufferSize(IMemAllocator *pAlloc,
                 ALLOCATOR_PROPERTIES *pProperties);
    HRESULT GetMediaType(int iPosition, CMediaType *pMediaType);
    // This function is called to process each new frame in the video.
    HRESULT Transform(IMediaSample *pIn, IMediaSample *pOut);
    HRESULT Transform(IMediaSample *pMediaSample);
    
    // Reveals the interface the filter support
    STDMETHODIMP NonDelegatingQueryInterface(REFIID riid, void ** ppv);
    // ISpecifyPropertyPages interface
    STDMETHODIMP GetPages(CAUUID *pPages);

// ISWFilter interface to custom the filter
    // Get the current parameters of the filter
    STDMETHODIMP GetParams(SWFilterParams *irp);
    // Set the parameters of the filter
    STDMETHODIMP SetParams(SWFilterParams *irp);

// CPersistStream stuff -- Save the config persistently
// These functions are used to serialize object, (like save or load
// graph file (*.grf) in GraphEdt.exe. It will be automaticly called 
// when you save filter to file.
    HRESULT WriteToStream(IStream *pStream);
    HRESULT ReadFromStream(IStream *pStream);
    int     SizeMax() { return 2 * STR_MAX_LENGTH; }
    // This is the only method of IPersist
    STDMETHODIMP GetClassID(CLSID *pClsid);

//my functions

	int		frameNr;							// Frame Nr.
	int		blackpixel;
	int		rotanteil;
    int		gruenanteil;
	int		blauanteil;
	//bool	Synchro;
	bool	firstframe;
	int		firstframehelligkeit;

	struct Zonenarray
	{
		int  Nr[50];							// Zonennummer
		float X[50];							// X-Koordinate (Schwerpunkt einer Zone)
		float Y[50];							// Y-Koordinate (Schwerpunkt einer Zone)
		float R[50];							// Rot-Anteil
		float G[50];							// Grn-Anteil
		float B[50];							// Blau-Anteil
		char  F[50][2];							// Maximale Farbe
		int  Gw[50];							// Gewichtung (Anzahl beteiligter Pixel an einer Zone)
	};	


	// Funktion zum Zeichnen eines Kreuzes ber dem Schwerpunkt einer Zone
	// Je grsser die Gewichtung desto grsser das Kreuz
	// Die Farbe des Kreuzes entspricht der Farbe der Zone
    void    drawCross(BYTE*       prgb,         
                      int         cxImage,
                      int         cyImage,
                      int         iPixelSize,
                      int         XCoord,
                      int         YCoord,
                      int         Gewicht,
                      char        Farbe);

	// rekursive Funktion fr die Suche der Zonen
	void 	suche   ( int		  cxImage,		// Bildbreite
					  int		  cyImage,		// Bildhhe
					  int		  aktZoneNr,	// aktuelle Zonen-Nr
					  BYTE*		  prgb,			// Zeiger auf aktuelle Position im Frame
					  int		  iPixelSize,	// Grsse eines Pixels
					  Zonenarray* Liste,		// Zeiger auf die ZonenListe
					  BYTE*		  prgbsave,		// Zeiger auf die Anfangsposition im Frame
					  int		  tiefe);


	// Maximale Farbe bestimmen
	// Rckgabewerte: R fr rot,  G fr grn,  B fr blau,  W fr weiss
	char    maxFarbe( float		  rot,			
					  float		  gruen,		
					  float		  blau);		


	// drei chars zu einem zusammenfgen
	char*   concat  ( char*		  a,			// Dateiname	
					  char*		  b,			// Laufnummer
					  char*		  c);			// Dateierweiterung

	
	
	//Declare NetworkClient Object
	ClientSocket* m_client;


protected:
    // Check if we can accepte the input media type
    BOOL    canPerformTransform(const CMediaType *pMediaType) const;
    // Copy from the input pin to output pin
    HRESULT copyMediaSample(IMediaSample *pSource, IMediaSample *pDest) const;
    // Save the configuration into registry
    void    saveConfig();
    // Read Config from registry
    void    readConfig(); 

    // Parameters controling the filter behavior
    SWFilterParams m_SWFilterParams;
    // Critical Section (used for multi-thread share)
    CCritSec m_SWFilterLock;

	//Stores First Frame for Background Compensation
	BYTE* _picFirstFrame;

	int* m_List;
	static const int _ZoneMAX = 40;
};
